import pytest
from UnleashClient.features import Feature
from UnleashClient.strategies import RemoteAddress, UserWithId, Default
from UnleashClient.variants import Variants
from tests.utilities import generate_email_list
from tests.utilities.testing_constants import IP_LIST
from tests.utilities.mocks.mock_variants import VARIANTS


(EMAIL_LIST, CONTEXT) = generate_email_list(20)


@pytest.fixture()
def test_feature():
    strategies = [RemoteAddress(parameters={"IPs": IP_LIST}), UserWithId(parameters={"userIds": EMAIL_LIST})]
    yield Feature("My Feature", True, strategies)


@pytest.fixture()
def test_feature_variants():
    strategies = [Default()]
    variants = Variants(VARIANTS, "My Feature")
    yield Feature("My Feature", True, strategies, variants)


def test_create_feature_true(test_feature):
    my_feature = test_feature

    CONTEXT["remoteAddress"] = "69.208.0.1"
    assert my_feature.is_enabled(CONTEXT)
    assert my_feature.yes_count == 1

    my_feature.reset_stats()
    assert my_feature.yes_count == 0


def test_create_feature_false(test_feature):
    my_feature = test_feature

    CONTEXT["remoteAddress"] = "1.208.0.1"
    CONTEXT["userId"] = "random@random.com"
    assert not my_feature.is_enabled(CONTEXT)
    assert my_feature.no_count == 1

    my_feature.reset_stats()
    assert my_feature.no_count == 0


def test_create_feature_not_enabled(test_feature):
    my_feature = test_feature
    my_feature.enabled = False

    CONTEXT["remoteAddress"] = "69.208.0.1"
    assert not my_feature.is_enabled(CONTEXT)


def test_create_feature_exception(test_feature):
    strategies = [{}, UserWithId(parameters={"userIds": EMAIL_LIST})]
    my_feature = Feature("My Feature", True, strategies)

    CONTEXT["remoteAddress"] = "69.208.0.1"
    assert not my_feature.is_enabled(CONTEXT)


def test_select_variation_novariation(test_feature):
    selected_variant = test_feature.get_variant()
    assert type(selected_variant) == dict
    assert selected_variant['name'] == 'disabled'


def test_select_variation_variation(test_feature_variants):
    selected_variant = test_feature_variants.get_variant({'userId': '2'})
    assert selected_variant['enabled']
    assert selected_variant['name'] == 'VarB'



import requests
from UnleashClient.constants import REQUEST_TIMEOUT, FEATURES_URL
from UnleashClient.utils import LOGGER


# pylint: disable=broad-except
def get_feature_toggles(url: str,
                        app_name: str,
                        instance_id: str,
                        custom_headers: dict,
                        custom_options: dict) -> dict:
    """
    Retrieves feature flags from unleash central server.

    Notes:
    * If unsuccessful (i.e. not HTTP status code 200), exception will be caught and logged.
      This is to allow "safe" error handling if unleash server goes down.

    :param url:
    :param app_name:
    :param instance_id:
    :param custom_headers:
    :param custom_options:
    :return: Feature flags if successful, empty dict if not.
    """
    try:
        LOGGER.info("Getting feature flag.")

        headers = {
            "UNLEASH-APPNAME": app_name,
            "UNLEASH-INSTANCEID": instance_id
        }

        resp = requests.get(url + FEATURES_URL,
                            headers={**custom_headers, **headers},
                            timeout=REQUEST_TIMEOUT, **custom_options)

        if resp.status_code != 200:
            LOGGER.warning("unleash feature fetch failed!")
            raise Exception("unleash feature fetch failed!")

        return resp.json()
    except Exception:
        LOGGER.exception("Unleash feature fetch failed!")

    return {}



from fcache.cache import FileCache
from UnleashClient.features.Feature import Feature
from UnleashClient.variants.Variants import Variants
from UnleashClient.constants import FEATURES_URL
from UnleashClient.utils import LOGGER


# pylint: disable=broad-except
def _create_strategies(provisioning: dict,
                       strategy_mapping: dict) -> list:
    feature_strategies = []

    for strategy in provisioning["strategies"]:
        try:
            if "parameters" in strategy.keys():
                strategy_provisioning = strategy['parameters']
            else:
                strategy_provisioning = {}

            if "constraints" in strategy.keys():
                constraint_provisioning = strategy['constraints']
            else:
                constraint_provisioning = {}

            feature_strategies.append(strategy_mapping[strategy['name']](constraints=constraint_provisioning, parameters=strategy_provisioning))
        except Exception as excep:
            LOGGER.warning("Failed to load strategy.  This may be a problem with a custom strategy.  Exception: %s",
                           excep)

    return feature_strategies


def _create_feature(provisioning: dict,
                    strategy_mapping: dict) -> Feature:
    if "strategies" in provisioning.keys():
        parsed_strategies = _create_strategies(provisioning, strategy_mapping)
    else:
        parsed_strategies = []

    if "variants" in provisioning:
        variant = Variants(provisioning['variants'], provisioning['name'])
    else:
        variant = None

    return Feature(name=provisioning["name"],
                   enabled=provisioning["enabled"],
                   strategies=parsed_strategies,
                   variants=variant
                   )


def load_features(cache: FileCache,
                  feature_toggles: dict,
                  strategy_mapping: dict) -> None:
    """
    Caching

    :param cache: Should be the cache class variable from UnleashClient
    :param feature_toggles: Should be the features class variable from UnleashClient
    :return:
    """
    # Pull raw provisioning from cache.
    try:
        feature_provisioning = cache[FEATURES_URL]

        # Parse provisioning
        parsed_features = {}
        feature_names = [d["name"] for d in feature_provisioning["features"]]

        for provisioning in feature_provisioning["features"]:
            parsed_features[provisioning["name"]] = provisioning

        # Delete old features/cache
        for feature in list(feature_toggles.keys()):
            if feature not in feature_names:
                del feature_toggles[feature]

        # Update existing objects
        for feature in feature_toggles.keys():
            feature_for_update = feature_toggles[feature]
            strategies = parsed_features[feature]["strategies"]

            feature_for_update.enabled = parsed_features[feature]["enabled"]
            if strategies:
                parsed_strategies = _create_strategies(parsed_features[feature], strategy_mapping)
                feature_for_update.strategies = parsed_strategies

            if 'variants' in parsed_features[feature]:
                feature_for_update.variants = Variants(
                    parsed_features[feature]['variants'],
                    parsed_features[feature]['name']
                )

        # Handle creation or deletions
        new_features = list(set(feature_names) - set(feature_toggles.keys()))

        for feature in new_features:
            feature_toggles[feature] = _create_feature(parsed_features[feature], strategy_mapping)
    except KeyError as cache_exception:
        LOGGER.warning("Cache Exception: %s", cache_exception)
        LOGGER.warning("Unleash client does not have cached features.  Please make sure client can communicate with Unleash server!")



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
   "version":1,
   "features":[
      {
         "name":"Feature.constraints.simple",
         "description":"Enabled only in dev",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[
                        "dev"
                     ]
                  }
               ]
            }
         ]
      },
      {
         "name":"Feature.constraints.list",
         "description":"Enabled in dev, stage, prod",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[
                        "dev",
                        "stage",
                        "prod"
                     ]
                  }
               ]
            }
         ]
      },
      {
         "name":"Feature.constraints.multi",
         "description":"Enabled in prod and only for users 1, 2, 3",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[
                        "prod"
                     ]
                  },
                  {
                     "contextName":"userId",
                     "operator":"IN",
                     "values":[
                        "1",
                        "2",
                        "3"
                     ]
                  },
                  {
                     "contextName":"appName",
                     "operator":"NOT_IN",
                     "values":[
                        "web",
                        "sun-app"
                     ]
                  }
               ]
            }
         ]
      },
      {
         "name":"Feature.constraints.custom",
         "description":"Enabled in prod in country IN (norway or sweeden)",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[
                        "prod"
                     ]
                  },
                  {
                     "contextName":"country",
                     "operator":"IN",
                     "values":[
                        "norway",
                        "sweeden"
                     ]
                  }
               ]
            }
         ]
      },
      {
         "name":"Feature.constraints.dual",
         "description":"Enabled in prod and not in qa and dev",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[
                        "prod",
                        "dev"
                     ]
                  },
                  {
                     "contextName":"environment",
                     "operator":"NOT_IN",
                     "values":[
                        "dev",
                        "qa"
                     ]
                  }
               ]
            }
         ]
      },
      {
         "name":"Feature.constraints.empty",
         "description":"Always disabled for empty list of values",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{

               },
               "constraints":[
                  {
                     "contextName":"environment",
                     "operator":"IN",
                     "values":[

                     ]
                  }
               ]
            }
         ]
      }
   ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_simple_defaultenv(unleash_client):
    """
    Feature.constraints.simple should not be enabled in default environment.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.constraints.simple", {})


@responses.activate
def test_feature_constraintssimple_devenv(unleash_client):
    """
    Feature.constraints.simple should be enabled in dev environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "dev"
    assert unleash_client.is_enabled("Feature.constraints.simple", {})


@responses.activate
def test_feature_constraintslist_stageenv(unleash_client):
    """
    Feature.constraints.list should be enabled in stage environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "stage"
    assert unleash_client.is_enabled("Feature.constraints.list", {})


@responses.activate
def test_feature_constraintslist_qaenv(unleash_client):
    """
    Feature.constraints.list should NOT be enabled in qa environment
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "qa"
    assert not unleash_client.is_enabled("Feature.constraints.list", {})


@responses.activate
def test_feature_constraintsmulti_user2(unleash_client):
    """
    Feature.constraints.multi should be enabled in prod for user 2
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    assert unleash_client.is_enabled("Feature.constraints.multi", {"userId": "2"})


@responses.activate
def test_feature_constraintsmulti_unknownuser(unleash_client):
    """
    Feature.constraints.multi should NOT be enabled for unknown user
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    assert not unleash_client.is_enabled("Feature.constraints.multi", {})


@responses.activate
def test_feature_constraintsmulti_web(unleash_client):
    """
    Feature.constraints.multi should NOT be enabled appName=web
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    unleash_client.unleash_static_context["appName"] = "web"
    assert not unleash_client.is_enabled("Feature.constraints.multi", {"userId": "2"})


@responses.activate
def test_feature_constraintscustom_norway(unleash_client):
    """
    Feature.constraints.custom should be enabled in prod for norway.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    unleash_client.unleash_static_context["appName"] = "web"
    assert unleash_client.is_enabled("Feature.constraints.custom", {"properties": {"country": "norway"}})


@responses.activate
def test_feature_constraintscustom_denmark(unleash_client):
    """
    Feature.constraints.custom should NOT be enabled in prod for denmark
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    unleash_client.unleash_static_context["appName"] = "web"
    assert not unleash_client.is_enabled("Feature.constraints.custom", {"properties": {"country": "denmark"}})


@responses.activate
def test_feature_constraintsdual_prod(unleash_client):
    """
    Feature.constraints.dual should be enabled in prod
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "prod"
    assert unleash_client.is_enabled("Feature.constraints.dual", {})


@responses.activate
def test_feature_constraintsdual_dev(unleash_client):
    """
    Feature.constraints.dual should not be enabled in dev
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "dev"
    assert not unleash_client.is_enabled("Feature.constraints.dual", {})


@responses.activate
def test_feature_constraintsempty(unleash_client):
    """
    Feature.constraints.empty should always be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    unleash_client.unleash_static_context["environment"] = "dev"
    assert not unleash_client.is_enabled("Feature.constraints.empty", {})



from UnleashClient.utils import LOGGER, get_identifier

class Constraint:
    def __init__(self, constraint_dict: dict) -> None:
        """
        Represents a constraint on a strategy

        constraint_dict = From the strategy document.
        """
        self.context_name = constraint_dict['contextName']
        self.operator = constraint_dict['operator']
        self.values = constraint_dict['values']

    def apply(self, context: dict = None) -> bool:
        """
        Returns true/false depending on constraint provisioning and context.

        :param context: Context information
        :return:
        """
        constraint_check = False

        try:
            value = get_identifier(self.context_name, context)

            if value:
                if self.operator.upper() == "IN":
                    constraint_check = value in self.values
                elif self.operator.upper() == "NOT_IN":
                    constraint_check = value not in self.values
        except Exception as excep:  #pylint: disable=W0703
            LOGGER.info("Could not evaluate context %s!  Error: %s", self.context_name, excep)

        return constraint_check



import random
import copy
from typing import Dict
from UnleashClient import utils
from UnleashClient.constants import DISABLED_VARIATION


class Variants():
    def __init__(self, variants_list: list, feature_name: str) -> None:
        """
        Represents an A/B test

        variants_list = From the strategy document.
        """
        self.variants = variants_list
        self.feature_name = feature_name

    def _apply_overrides(self, context: dict) -> dict:
        """
        Figures out if an override should be applied based on a context.

        Notes:
            - This matches only the first variant found.
        """
        variants_with_overrides = [x for x in self.variants if 'overrides' in x.keys()]
        override_variant = {}  # type: Dict

        for variant in variants_with_overrides:
            for override in variant['overrides']:
                identifier = utils.get_identifier(override['contextName'], context)
                if identifier in override["values"]:
                    override_variant = variant

        return override_variant

    @staticmethod
    def _get_seed(context: dict) -> str:
        """
        Grabs seed value from context.
        """
        seed = str(random.random() * 10000)

        if 'userId' in context:
            seed = context['userId']
        elif 'sessionId' in context:
            seed = context['sessionId']
        elif 'remoteAddress' in context:
            seed = context['remoteAddress']

        return seed

    @staticmethod
    def _format_variation(variation: dict) -> dict:
        formatted_variation = copy.deepcopy(variation)
        del formatted_variation['weight']
        if 'overrides' in formatted_variation:
            del formatted_variation['overrides']
        return formatted_variation

    def get_variant(self, context: dict) -> dict:
        """
        Determines what variation a user is in.

        :param context:
        :return:
        """
        fallback_variant = copy.deepcopy(DISABLED_VARIATION)

        if self.variants:
            override_variant = self._apply_overrides(context)
            if override_variant:
                return self._format_variation(override_variant)

            total_weight = sum([x['weight'] for x in self.variants])
            if total_weight <= 0:
                return fallback_variant

            target = utils.normalized_hash(self._get_seed(context), self.feature_name, total_weight)
            counter = 0
            for variation in self.variants:
                counter += variation['weight']

                if counter >= target:
                    return self._format_variation(variation)

        # Catch all return.
        return fallback_variant



import pytest
from UnleashClient.constraints import Constraint


CONSTRAINT_DICT_IN = \
    {
        "contextName": "appName",
        "operator": "IN",
        "values": [
            "test",
            "test2"
        ]
    }


CONSTRAINT_DICT_NOTIN = \
    {
        "contextName": "appName",
        "operator": "NOT_IN",
        "values": [
            "test",
            "test2"
        ]
    }


@pytest.fixture()
def constraint_IN():
    yield Constraint(CONSTRAINT_DICT_IN)


@pytest.fixture()
def constraint_NOTIN():
    yield Constraint(CONSTRAINT_DICT_NOTIN)


def test_constraint_IN_match(constraint_IN):
    constraint = constraint_IN
    context = {
        'appName': 'test'
    }

    assert constraint.apply(context)


def test_constraint_IN_not_match(constraint_IN):
    constraint = constraint_IN
    context = {
        'appName': 'test3'
    }

    assert not constraint.apply(context)


def test_constraint_IN_missingcontext(constraint_IN):
    constraint = constraint_IN
    assert not constraint.apply({})


def test_constraint_NOTIN_match(constraint_NOTIN):
    constraint = constraint_NOTIN
    context = {
        'appName': 'test'
    }

    assert not constraint.apply(context)


def test_constraint_NOTIN_not_match(constraint_NOTIN):
    constraint = constraint_NOTIN
    context = {
        'appName': 'test3'
    }

    assert constraint.apply(context)



# Library
SDK_NAME = "unleash-client-python"
SDK_VERSION = "3.4.0"
REQUEST_TIMEOUT = 30
METRIC_LAST_SENT_TIME = "mlst"

# =Unleash=
APPLICATION_HEADERS = {"Content-Type": "application/json"}
DISABLED_VARIATION = {
        'name': 'disabled',
        'enabled': False
}

# Paths
REGISTER_URL = "/client/register"
FEATURES_URL = "/client/features"
METRICS_URL = "/client/metrics"



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [{
      "name": "Feature.A5",
      "description": "Enabled toggle for 100%",
      "enabled": true,
      "strategies": [{
        "name": "gradualRolloutRandom",
        "parameters": {
          "percentage": "100"
        }
      }]
    },
    {
      "name": "Feature.B5",
      "description": "Disabled toggle with 0% rollout",
      "enabled": true,
      "strategies": [{
        "name": "gradualRolloutRandom",
        "parameters": {
          "percentage": "0"
        }
      }]
    },
    {
      "name": "Feature.C5",
      "enabled": true,
      "strategies": [{
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": "0"
          }
        },
        {
          "name": "default"
        }
      ]
    },
    {
      "name": "Feature.D5",
      "description": "Disabled toggle should be disabled",
      "enabled": false,
      "strategies": [{
        "name": "gradualRolloutRandom",
        "parameters": {
          "percentage": "100"
        }
      }]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_a5(unleash_client):
    """
    Feature.A5 should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.A5")


@responses.activate
def test_feature_b5(unleash_client):
    """
    Feature.B5 should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.B5")


@responses.activate
def test_feature_c5(unleash_client):
    """
    Feature.C5 should be enabled because of default strategy
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.C5")


@responses.activate
def test_feature_d5(unleash_client):
    """
    Feature.D5 should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.D5")



import logging
import mmh3  # pylint: disable=import-error

LOGGER = logging.getLogger(__name__)

def normalized_hash(identifier: str,
                    activation_group: str,
                    normalizer: int = 100) -> int:
    return mmh3.hash("{}:{}".format(activation_group, identifier), signed=False) % normalizer + 1


def get_identifier(context_key_name: str, context: dict) -> str:
    if context_key_name in context.keys():
        value = context[context_key_name]
    elif 'properties' in context.keys() and context_key_name in context['properties'].keys():
        value = context['properties'][context_key_name]
    else:
        value = None

    return value



MOCK_FEATURE_RESPONSE = {
    "version": 1,
    "features": [
        {
            "name": "testFlag",
            "description": "This is a test!",
            "enabled": True,
            "strategies": [
                {
                    "name": "default",
                    "parameters": {}
                }
            ],
            "createdAt": "2018-10-04T01:27:28.477Z"
        },
        {
            "name": "testFlag2",
            "description": "Test flag 2",
            "enabled": True,
            "strategies": [
                {
                    "name": "gradualRolloutRandom",
                    "parameters": {
                        "percentage": 50
                    }
                }
            ],
            "createdAt": "2018-10-04T11:03:56.062Z"
        },
        {
            "name": "testContextFlag",
            "description": "This is a test for static context fileds!",
            "enabled": True,
            "strategies": [
                {
                    "name": "custom-context",
                    "parameters": {
                        "environments": "prod"
                    }
                }
            ],
            "createdAt": "2018-10-04T01:27:28.477Z"
        },
        {
          "name": "testVariations",
          "description": "Test variation",
          "enabled": True,
          "strategies": [
              {
                  "name": "userWithId",
                  "parameters": {
                      "userIds": "2"
                  }
              }
          ],
          "variants": [
              {
                  "name": "VarA",
                  "weight": 34,
                  "payload": {
                      "type": "string",
                      "value": "Test1"
                  },
                  "overrides": [
                      {
                          "contextName": "userId",
                          "values": [
                              "ivanklee86@gmail.com",
                              "ivan@aaptiv.com"
                          ]
                      }
                  ]
              },
              {
                  "name": "VarB",
                  "weight": 33,
                  "payload": {
                      "type": "string",
                      "value": "Test 2"
                  }
              },
              {
                  "name": "VarC",
                  "weight": 33,
                  "payload": {
                      "type": "string",
                      "value": "Test 3"
                  }
              }
          ],
          "createdAt": "2019-10-25T13:22:02.035Z"
      }
    ]
}



from collections import ChainMap
from datetime import datetime, timezone
import fcache
from UnleashClient.api import send_metrics
from UnleashClient.constants import METRIC_LAST_SENT_TIME


def aggregate_and_send_metrics(url: str,
                               app_name: str,
                               instance_id: str,
                               custom_headers: dict,
                               custom_options: dict,
                               features: dict,
                               ondisk_cache: fcache.cache
                               ) -> None:
    feature_stats_list = []

    for feature_name in features.keys():
        feature_stats = {
            features[feature_name].name: {
                "yes": features[feature_name].yes_count,
                "no": features[feature_name].no_count
            }
        }

        features[feature_name].reset_stats()
        feature_stats_list.append(feature_stats)

    metrics_request = {
        "appName": app_name,
        "instanceId": instance_id,
        "bucket": {
            "start": ondisk_cache[METRIC_LAST_SENT_TIME].isoformat(),
            "stop": datetime.now(timezone.utc).isoformat(),
            "toggles": dict(ChainMap(*feature_stats_list))
        }
    }

    send_metrics(url, metrics_request, custom_headers, custom_options)
    ondisk_cache[METRIC_LAST_SENT_TIME] = datetime.now(timezone.utc)
    ondisk_cache.sync()



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [{
      "name": "Feature.remoteAddress.A",
      "description": "Enabled toggle for localhost",
      "enabled": true,
      "strategies": [{
        "name": "remoteAddress",
        "parameters": {
          "IPs": "127.0.0.1"
        }
      }]
    },
    {
      "name": "Feature.remoteAddress.B",
      "description": "Enabled toggle for list of IPs",
      "enabled": true,
      "strategies": [{
        "name": "remoteAddress",
        "parameters": {
          "IPs": "192.168.0.1, 192.168.0.2, 192.168.0.3"
        }
      }]
    },
    {
      "name": "Feature.remoteAddress.C",
      "description": "Ignore invalid IP's in list",
      "enabled": true,
      "strategies": [{
        "name": "remoteAddress",
        "parameters": {
          "IPs": "192.168.0.1, 192.invalid, 192.168.0.2, 192.168.0.3"
        }
      }]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_remoteaddress_a_enabled(unleash_client):
    """
    Feature.remoteAddress.A should be enabled for localhost
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "remoteAddress": "127.0.0.1"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.remoteAddress.A", context_values)


@responses.activate
def test_feature_remoteaddress_a_nocontext(unleash_client):
    """
    Feature.remoteAddress.A should not be enabled for missing remoteAddress on context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {}

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.remoteAddress.A", context_values)


@responses.activate
def test_feature_remoteaddress_b_enabled(unleash_client):
    """
    Feature.remoteAddress.B should be enabled for remoteAddress 192.168.0.1
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "remoteAddress": "192.168.0.1"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.remoteAddress.B", context_values)


@responses.activate
def test_feature_remoteaddress_b_list(unleash_client):
    """
    Feature.remoteAddress.B should be enabled for remoteAddress 192.168.0.3
    TODO: Error in spec.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "remoteAddress": "192.168.0.3"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.remoteAddress.B", context_values)


@responses.activate
def test_feature_remoteaddress_b_notlist(unleash_client):
    """
    Feature.remoteAddress.B should not be enabled for remoteAddress 217.100.10.11
    TODO: Error in spec.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "remoteAddress": "217.100.10.11"
    }

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.remoteAddress.B", context_values)


@responses.activate
def test_feature_remoteaddress_c(unleash_client):
    """
    Feature.remoteAddress.C should be enabled for remoteAddress 192.168.0.3
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "remoteAddress": "192.168.0.3"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.remoteAddress.C", context_values)



# pylint: disable=dangerous-default-value
import warnings
from UnleashClient.constraints import Constraint


class Strategy:
    """
    The parent class for default and custom strategies.

    In general, default & custom classes should only need to override:
    * __init__() - Depending on the parameters your feature needs
    * apply() - Your feature provisioning
    """
    def __init__(self,
                 constraints: list = [],
                 parameters: dict = {},
                 ) -> None:
        """
        A generic strategy objects.

        :param constraints: List of 'constraints' objects derived from strategy section (...from feature section) of
        /api/clients/features response
        :param parameters: The 'parameter' objects from the strategy section (...from feature section) of
        /api/clients/features response
        """
        self.parameters = parameters
        self.constraints = constraints
        self.parsed_constraints = self.load_constraints(constraints)
        self.parsed_provisioning = self.load_provisioning()

    def __call__(self, context: dict = None):
        warnings.warn(
            "unleash-client-python v3.x.x requires overriding the execute() method instead of the __call__() method.",
            DeprecationWarning
        )

    def execute(self, context: dict = None) -> bool:
        """
        Executes the strategies by:
        - Checking constraints
        - Applying the strategy

        :param context: Context information
        :return:
        """
        flag_state = False

        if all([constraint.apply(context) for constraint in self.parsed_constraints]):
            flag_state = self.apply(context)

        return flag_state

    def load_constraints(self, constraints_list: list) -> list:  #pylint: disable=R0201
        """
        Loads constraints from provisioning.

        :return:
        """
        parsed_constraints_list = []

        for constraint_dict in constraints_list:
            parsed_constraints_list.append(Constraint(constraint_dict=constraint_dict))

        return parsed_constraints_list

    # pylint: disable=no-self-use
    def load_provisioning(self) -> list:
        """
        Method to load data on object initialization, if desired.

        This should parse the raw values in self.parameters into format Python can comprehend.
        """
        return []

    def apply(self, context: dict = None) -> bool:  #pylint: disable=W0613,R0201
        """
        Strategy implementation goes here.

        :param context:
        :return:
        """
        return False



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [{
      "name": "Feature.multiStrategies.A",
      "description": "Enabled for via last stratgy",
      "enabled": true,
      "strategies": [{
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": "0"
          }
        },
        {
          "name": "default",
          "parameters": {}
        }
      ]
    },
    {
      "name": "Feature.multiStrategies.B",
      "description": "Enabled for user=123",
      "enabled": true,
      "strategies": [{
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": "0"
          }
        },
        {
          "name": "userWithId",
          "parameters": {
            "userIds": "123"
          }
        }
      ]
    },
    {
      "name": "Feature.multiStrategies.C",
      "description": "Enabled for user=123",
      "enabled": true,
      "strategies": [{
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": "0"
          }
        },
        {
          "name": "userWithId",
          "parameters": {
            "userIds": "123"
          }
        },
        {
          "name": "gradualRolloutRandom",
          "parameters": {
            "percentage": "0"
          }
        }
      ]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_multiStrategies_a_enabled(unleash_client):
    """
    Feature.multiStrategies.A should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.multiStrategies.A")


@responses.activate
def test_feature_multiStrategies_b_nouser(unleash_client):
    """
    Feature.multiStrategies.B disabled for unknown user
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context = {}

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.multiStrategies.B", context)


@responses.activate
def test_feature_multiStrategies_b_enabled(unleash_client):
    """
    Feature.multiStrategies.B enabled for user=123
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context = {
        "userId": "123"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.multiStrategies.B", context)


@responses.activate
def test_feature_multiStrategies_c_unknown(unleash_client):
    """
    Feature.multiStrategies.C disabled for unknown users
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context = {}

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.multiStrategies.C", context)


@responses.activate
def test_feature_multiStrategies_c_disabled(unleash_client):
    """
    Feature.multiStrategies.C disabled for user=22
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context = {
        "userId": "22"
    }

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.multiStrategies.C", context)


@responses.activate
def test_feature_multiStrategies_c_enabled(unleash_client):
    """
    Feature.multiStrategies.C enabled for user=123
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context = {
        "userId": "123"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.multiStrategies.C", context)



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
  "version": 1,
  "features": [{
      "name": "Feature.A2",
      "description": "Enabled toggle",
      "enabled": true,
      "strategies": [{
        "name": "userWithId",
        "parameters": {
          "userIds": "123"
        }
      }]
    },
    {
      "name": "Feature.B2",
      "description": "Disabled toggle",
      "enabled": true,
      "strategies": [{
        "name": "userWithId",
        "parameters": {
          "userIds": "123"
        }
      }]
    },
    {
      "name": "Feature.C2",
      "enabled": true,
      "strategies": [{
          "name": "userWithId",
          "parameters": {
            "userIds": "123"
          }
        },
        {
          "name": "default"
        }
      ]
    },
    {
      "name": "Feature.D2",
      "enabled": true,
      "strategies": [{
        "name": "userWithId",
        "parameters": {
          "userIds": "123, 222, 88"
        }
      }]
    }
  ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_a2_enabled(unleash_client):
    """
    Feature.A2 Should be enabled for user on context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "userId": "123"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.A2", context_values)


@responses.activate
def test_feature_a2_disabled(unleash_client):
    """
    Feature.A2 Should not be enabled for user not in context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "userId": "22"
    }

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.A2", context_values)


@responses.activate
def test_feature_c2(unleash_client):
    """
    Feature.C2 Should not be "disabled" for for everyone
    TODO: Contact team about issue in test case description, should be enabled.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "userId": "22"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.C2", context_values)


@responses.activate
def test_feature_d2(unleash_client):
    """
    Feature.D2 Should "-not-" be enabled for user in list
    TODO: Contact team about issue in test case description, should be enabled.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "userId": "222"
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.D2", context_values)


@responses.activate
def test_feature_no_id(unleash_client):
    """
    Feature.A2 Should be disabled when no userId on context
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {}

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.A2", context_values)



import copy
from UnleashClient.loader import load_features
from UnleashClient.features import Feature
from UnleashClient.strategies import GradualRolloutUserId, FlexibleRollout, UserWithId
from UnleashClient.variants import Variants
from UnleashClient.constants import FEATURES_URL
from tests.utilities.mocks import MOCK_ALL_FEATURES
from tests.utilities.testing_constants import DEFAULT_STRATEGY_MAPPING
from tests.utilities.decorators import cache_full, cache_custom  # noqa: F401

MOCK_UPDATED = copy.deepcopy(MOCK_ALL_FEATURES)
MOCK_UPDATED["features"][4]["strategies"][0]["parameters"]["percentage"] = 60


def test_loader_initialization(cache_full):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_full

    # Tests
    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)
    assert isinstance(in_memory_features["GradualRolloutUserID"], Feature)
    assert isinstance(in_memory_features["GradualRolloutUserID"].strategies[0], GradualRolloutUserId)

    for feature_name in in_memory_features.keys():
        feature = in_memory_features[feature_name]
        assert len(feature.strategies) > 0
        strategy = feature.strategies[0]

        if isinstance(strategy, UserWithId):
            assert strategy.parameters
            assert len(strategy.parsed_provisioning)

        if isinstance(strategy, FlexibleRollout):
            len(strategy.parsed_constraints) > 0

        if isinstance(strategy, Variants):
            assert strategy.variants


def test_loader_refresh(cache_full):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_full

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    # Simulate update mutation
    temp_cache[FEATURES_URL] = MOCK_UPDATED
    temp_cache.sync()

    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)

    assert in_memory_features["GradualRolloutUserID"].strategies[0].parameters["percentage"] == 60


def test_loader_initialization_failure(cache_custom):  # noqa: F811
    # Set up variables
    in_memory_features = {}
    temp_cache = cache_custom

    # Tests
    load_features(temp_cache, in_memory_features, DEFAULT_STRATEGY_MAPPING)
    assert isinstance(in_memory_features["UserWithId"], Feature)



import responses
from pytest import mark, param
from requests import ConnectionError
from tests.utilities.testing_constants import URL, CUSTOM_HEADERS, CUSTOM_OPTIONS
from tests.utilities.mocks.mock_metrics import MOCK_METRICS_REQUEST
from UnleashClient.constants import METRICS_URL
from UnleashClient.api import send_metrics


FULL_METRICS_URL = URL + METRICS_URL


@responses.activate
@mark.parametrize("payload,status,expected", (
    param({"json": {}}, 202, lambda result: result, id="success"),
    param({"json": {}}, 500, lambda result: not result, id="failure"),
    param({"body": ConnectionError("Test connection error.")}, 200, lambda result: not result, id="exception"),
))
def test_send_metrics(payload, status, expected):
    responses.add(responses.POST, FULL_METRICS_URL, **payload, status=status)

    result = send_metrics(URL, MOCK_METRICS_REQUEST, CUSTOM_HEADERS, CUSTOM_OPTIONS)

    assert len(responses.calls) == 1
    assert expected(result)



from .Strategy import Strategy
from .Default import Default
from .UserWithId import UserWithId
from .GradualRolloutUserId import GradualRolloutUserId
from .GradualRolloutSessionId import GradualRolloutSessionId
from .GradualRolloutRandom import GradualRolloutRandom
from .RemoteAddress import RemoteAddress
from .ApplicationHostname import ApplicationHostname
from .FlexibleRolloutStrategy import FlexibleRollout



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
    "version": 1,
    "features": [{
            "name": "Feature.A",
            "description": "Enabled toggle",
            "enabled": true,
            "strategies": [{
                "name": "default"
            }]
        },
        {
            "name": "Feature.B",
            "description": "Disabled toggle",
            "enabled": false,
            "strategies": [{
                "name": "default"
            }]
        },
        {
            "name": "Feature.C",
            "enabled": true,
            "strategies": []
        }
    ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_a(unleash_client):
    """
    Feature.A should be enabled.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.A")


@responses.activate
def test_feature_b(unleash_client):
    """
    Feature.B should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Feature.B")


@responses.activate
def test_feature_c(unleash_client):
    """
    Feature.C should be enabled when strategy missing
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.C")


@responses.activate
def test_feature_unknown(unleash_client):
    """
    Unknown feature toggle should be disabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("Unknown")


@responses.activate
def test_feature_all_context_values(unleash_client):
    """
    Should allow all context values
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    context_values = {
        "userId": "123",
        "sessionId": "asd123",
        "remoteAddress": "127.0.0.1",
        "properties": {
            "customName": "customValue",
            "anotherName": "anotherValue"
        }
    }

    unleash_client.initialize_client()
    assert unleash_client.is_enabled("Feature.A", context_values)



import platform
from UnleashClient.strategies.Strategy import Strategy


class ApplicationHostname(Strategy):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["hostNames"].split(',')]

    def apply(self, context: dict = None) -> bool:
        """
        Returns true if userId is a member of id list.

        :return:
        """
        return platform.node() in self.parsed_provisioning



import mimesis


def generate_context():
    return {"userId": mimesis.Person('en').email()}


def generate_email_list(num: int) -> (str, dict):
    """
    Generates an unleash-style list of emails for testing.

    :param num:
    :return:
    """
    first_email = mimesis.Person('en').email()
    email_list_string = first_email

    context = {"userId": first_email}

    for _ in range(num - 1):
        email_list_string += "," + mimesis.Person('en').email()

    return (email_list_string, context)



import responses
from UnleashClient import UnleashClient
from UnleashClient.strategies import Strategy
from tests.utilities.testing_constants import URL, APP_NAME
from tests.utilities.mocks import MOCK_CUSTOM_STRATEGY
from tests.utilities.old_code.StrategyV2 import StrategyOldV2
from UnleashClient.constants import REGISTER_URL, FEATURES_URL, METRICS_URL


class CatTest(Strategy):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["sound"].split(',')]

    def apply(self, context: dict = None) -> bool:
        """
        Turn on if I'm a cat.

        :return:
        """
        default_value = False

        if "sound" in context.keys():
            default_value = context["sound"] in self.parsed_provisioning

        return default_value


class DogTest(StrategyOldV2):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["sound"].split(',')]

    def _call_(self, context: dict = None) -> bool:
        """
        Turn on if I'm a dog.

        :return:
        """
        default_value = False

        if "sound" in context.keys():
            default_value = context["sound"] in self.parsed_provisioning

        return default_value


@responses.activate
def test_uc_customstrategy_happypath(recwarn):
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_CUSTOM_STRATEGY, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    custom_strategies_dict = {
        "amIACat": CatTest,
        "amIADog": DogTest
    }

    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        environment="prod",
        custom_strategies=custom_strategies_dict)

    unleash_client.initialize_client()

    # Check custom strategy.
    assert unleash_client.is_enabled("CustomToggle", {"sound": "meow"})
    assert not unleash_client.is_enabled("CustomToggle", {"sound": "bark"})

    # Check warning on deprecated strategy.
    assert len(recwarn) == 1
    assert recwarn.pop(DeprecationWarning)


@responses.activate
def test_uc_customstrategy_depredationwarning():
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_CUSTOM_STRATEGY, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    custom_strategies_dict = {
        "amIACat": CatTest,
        "amIADog": DogTest
    }

    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        environment="prod",
        custom_strategies=custom_strategies_dict)

    unleash_client.initialize_client()

    # Check a toggle that contains an outdated custom strategy
    assert unleash_client.is_enabled("CustomToggleWarning", {"sound": "meow"})


@responses.activate
def test_uc_customstrategy_safemulti():
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_CUSTOM_STRATEGY, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    custom_strategies_dict = {
        "amIACat": CatTest,
        "amIADog": DogTest
    }

    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        environment="prod",
        custom_strategies=custom_strategies_dict)

    unleash_client.initialize_client()

    # Check a toggle that contains an outdated custom strategy and a default strategy.
    assert unleash_client.is_enabled("CustomToggleWarningMultiStrat", {"sound": "meow"})



import uuid
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.constants import FEATURES_URL, METRICS_URL, REGISTER_URL
from tests.utilities.testing_constants import URL, APP_NAME


MOCK_JSON = """
{
   "version":1,
   "features":[
      {
         "name":"Feature.Variants.A",
         "description":"Enabled",
         "enabled":true,
         "strategies":[
            {
               "name":"default",
               "parameters":{}
            }
         ],
         "variants":[
            {
               "name":"variant1",
               "weight":1,
               "payload":{
                  "type":"string",
                  "value":"val1"
               }
            }
         ]
      },
      {
         "name":"Feature.Variants.B",
         "description":"Enabled for user=123",
         "enabled":true,
         "strategies":[],
         "variants":[
            {
               "name":"variant1",
               "weight":1,
               "payload":{
                  "type":"string",
                  "value":"val1"
               }
            },
            {
               "name":"variant2",
               "weight":1,
               "payload":{
                  "type":"string",
                  "value":"val2"
               }
            }
         ]
      },
      {
         "name":"Feature.Variants.C",
         "description":"Testing three variants",
         "enabled":true,
         "strategies":[],
         "variants":[
            {
               "name":"variant1",
               "weight":33,
               "payload":{
                  "type":"string",
                  "value":"val1"
               }
            },
            {
               "name":"variant2",
               "weight":33,
               "payload":{
                  "type":"string",
                  "value":"val2"
               }
            },
            {
               "name":"variant3",
               "weight":33,
               "payload":{
                  "type":"string",
                  "value":"val3"
               }
            }
         ]
      },
      {
         "name":"Feature.Variants.D",
         "description":"Variants with payload",
         "enabled":true,
         "strategies":[],
         "variants":[
            {
               "name":"variant1",
               "weight":1,
               "payload":{
                  "type":"string",
                  "value":"val1"
               }
            },
            {
               "name":"variant2",
               "weight":49,
               "payload":{
                  "type":"string",
                  "value":"val2"
               }
            },
            {
               "name":"variant3",
               "weight":50,
               "payload":{
                  "type":"string",
                  "value":"val3"
               }
            }
         ]
      },
      {
         "name":"Feature.Variants.override.D",
         "description":"Variant with overrides",
         "enabled":true,
         "strategies":[],
         "variants":[
            {
               "name":"variant1",
               "weight":33,
               "payload":{
                  "type":"string",
                  "value":"val1"
               },
               "overrides":[
                  {
                     "contextName":"userId",
                     "values":[
                        "132",
                        "61"
                     ]
                  }
               ]
            },
            {
               "name":"variant2",
               "weight":33,
               "payload":{
                  "type":"string",
                  "value":"val2"
               }
            },
            {
               "name":"variant3",
               "weight":34,
               "payload":{
                  "type":"string",
                  "value":"val3"
               }
            }
         ]
      }
   ]
}
"""


@pytest.fixture()
def unleash_client():
    unleash_client = UnleashClient(url=URL,
                                   app_name=APP_NAME,
                                   instance_id='pytest_%s' % uuid.uuid4())
    yield unleash_client
    unleash_client.destroy()


@responses.activate
def test_feature_variantsA_enabled(unleash_client):
    """
    Feature.Variants.A should be enabled
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "0"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.A", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsMissingToggle(unleash_client):
    """
    Feature.Variants.MissingToggle should be disabled missing toggle
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {}

    expected_result = {
        "name": "disabled",
        "enabled": False
    }

    actual_result = unleash_client.get_variant("Feature.Variants.MissingToggle", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsB_2(unleash_client):
    """
    Feature.Variants.B should be enabled for user 2.
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "2"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.B", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsB_0(unleash_client):
    """
    Feature.Variants.B should be enabled for user 0
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "0"
    }

    expected_result = {
        "name": "variant2",
        "payload": {
            "type": "string",
            "value": "val2"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.B", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsC_315(unleash_client):
    """
    Feature.Variants.C should return variant1 for user 315
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "315"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.C", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsC_320(unleash_client):
    """
    Feature.Variants.C should return variant1 for user 320
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "320"
    }

    expected_result = {
        "name": "variant2",
        "payload": {
            "type": "string",
            "value": "val2"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.C", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsC_729(unleash_client):
    """
    Feature.Variants.C should return variant1 for user 729
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "729"
    }

    expected_result = {
        "name": "variant3",
        "payload": {
            "type": "string",
            "value": "val3"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.C", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsD_enabled367(unleash_client):
    """
    Feature.Variants.D should return variant1 for user 367
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "367"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.D", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsD_enabled31(unleash_client):
    """
    Feature.Variants.D should return variant2 for user id 31
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "31"
    }

    expected_result = {
        "name": "variant2",
        "payload": {
            "type": "string",
            "value": "val2"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.D", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsD_19(unleash_client):
    """
    Feature.Variants.D should return variant3 for user ID 19
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "19"
    }

    expected_result = {
        "name": "variant3",
        "payload": {
            "type": "string",
            "value": "val3"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.D", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsoverrideD_132(unleash_client):
    """
    Feature.Variants.override.D should return variant1 for user ID 132
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "132"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.override.D", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsoverrideD_61(unleash_client):
    """
    Feature.Variants.override.D should return variant1 for user ID 61
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "61"
    }

    expected_result = {
        "name": "variant1",
        "payload": {
            "type": "string",
            "value": "val1"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.override.D", context)
    assert actual_result == expected_result


@responses.activate
def test_feature_variantsoverrideD_82(unleash_client):
    """
    Feature.Variants.override.D should return variant2 for user ID 82
    """
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=json.loads(MOCK_JSON), status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Tests
    unleash_client.initialize_client()
    context = {
        'userId': "82"
    }

    expected_result = {
        "name": "variant2",
        "payload": {
            "type": "string",
            "value": "val2"
        },
        "enabled": True
    }

    actual_result = unleash_client.get_variant("Feature.Variants.override.D", context)
    assert actual_result == expected_result



"""Setup file for UnleashClient"""
from setuptools import setup, find_packages


def readme():
    """Include README.rst content in PyPi build information"""
    with open('README.md') as file:
        return file.read()


setup(
    name='UnleashClient',
    version='3.4.0',
    author='Ivan Lee',
    author_email='ivanklee86@gmail.com',
    description='Python client for the Unleash feature toggle system!',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/Unleash/unleash-client-python',
    packages=find_packages(),
    install_requires=["requests==2.22.0",
                      "fcache==0.4.7",
                      "mmh3==2.5.1",
                      "apscheduler==3.6.3"],
    tests_require=['pytest', "mimesis", "responses", 'pytest-mock'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ]
)



#Old Strategy object from unleash-client-python version 1.x.x and 2.x.x

# pylint: disable=dangerous-default-value
class StrategyOldV2():
    """
    In general, default & custom classes should only need to override:
    * __call__() - Implementation of the strategy.
    * load_provisioning - Loads strategy provisioning
    """
    def __init__(self,
                 parameters: dict = {}) -> None:
        """
        A generic strategy objects.
        :param parameters: 'parameters' key from strategy section (...from feature section) of
        /api/clients/features response
        """
        # Experiment information
        self.parameters = parameters

        self.parsed_provisioning = self.load_provisioning()

    # pylint: disable=no-self-use
    def load_provisioning(self) -> list:
        """
        Method to load data on object initialization, if desired.
        This should parse the raw values in self.parameters into format Python can comprehend.
        """
        return []

    def __eq__(self, other):
        return self.parameters == other.parameters

    def __call__(self, context: dict = None) -> bool:
        """
        Strategy implementation goes here.
        :param context: Context information
        :return:
        """
        return False



import responses
from pytest import mark, param
from tests.utilities.mocks.mock_features import MOCK_FEATURE_RESPONSE
from tests.utilities.testing_constants import URL, APP_NAME, INSTANCE_ID, CUSTOM_HEADERS, CUSTOM_OPTIONS
from UnleashClient.constants import FEATURES_URL
from UnleashClient.api import get_feature_toggles


FULL_FEATURE_URL = URL + FEATURES_URL


@responses.activate
@mark.parametrize("response,status,expected", (
    param(MOCK_FEATURE_RESPONSE, 200, lambda result: result["version"] == 1, id="success"),
    param({}, 500, lambda result: not result, id="failure"),
))
def test_get_feature_toggle(response, status, expected):
    responses.add(responses.GET, FULL_FEATURE_URL, json=response, status=status)

    result = get_feature_toggles(URL,
                                 APP_NAME,
                                 INSTANCE_ID,
                                 CUSTOM_HEADERS,
                                 CUSTOM_OPTIONS)

    assert len(responses.calls) == 1
    assert expected(result)



import random
from UnleashClient.strategies.Strategy import Strategy
from UnleashClient.utils import normalized_hash


class FlexibleRollout(Strategy):
    @staticmethod
    def random_hash() -> int:
        return random.randint(1, 100)

    def apply(self, context: dict = None) -> bool:
        """
        If constraints are satisfied, return a percentage rollout on provisioned.

        :return:
        """
        percentage = int(self.parameters['rollout'])
        activation_group = self.parameters['groupId']
        stickiness = self.parameters['stickiness']

        if stickiness == 'default':
            if 'userId' in context.keys():
                calculated_percentage = normalized_hash(context['userId'], activation_group)
            elif 'sessionId' in context.keys():
                calculated_percentage = normalized_hash(context['sessionId'], activation_group)
            else:
                calculated_percentage = self.random_hash()
        elif stickiness in ['userId', 'sessionId']:
            calculated_percentage = normalized_hash(context[stickiness], activation_group)
        else:
            # This also handles the stickiness == random scenario.
            calculated_percentage = self.random_hash()

        return percentage > 0 and calculated_percentage <= percentage



from fcache.cache import FileCache
from UnleashClient.api import get_feature_toggles
from UnleashClient.loader import load_features
from UnleashClient.constants import FEATURES_URL
from UnleashClient.utils import LOGGER


def fetch_and_load_features(url: str,
                            app_name: str,
                            instance_id: str,
                            custom_headers: dict,
                            custom_options: dict,
                            cache: FileCache,
                            features: dict,
                            strategy_mapping: dict) -> None:
    feature_provisioning = get_feature_toggles(url, app_name, instance_id, custom_headers, custom_options)

    if feature_provisioning:
        cache[FEATURES_URL] = feature_provisioning
        cache.sync()
    else:
        LOGGER.warning("Unable to get feature flag toggles, using cached provisioning.")

    load_features(cache, features, strategy_mapping)



from .mock_metrics import MOCK_METRICS_REQUEST
from .mock_all_features import MOCK_ALL_FEATURES
from .mock_features import MOCK_FEATURE_RESPONSE
from .mock_custom_strategy import MOCK_CUSTOM_STRATEGY



from datetime import datetime, timezone
from typing import Dict, Callable, Any
import copy
from fcache.cache import FileCache
from apscheduler.job import Job
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger
from UnleashClient.api import register_client
from UnleashClient.periodic_tasks import fetch_and_load_features, aggregate_and_send_metrics
from UnleashClient.strategies import ApplicationHostname, Default, GradualRolloutRandom, \
    GradualRolloutSessionId, GradualRolloutUserId, UserWithId, RemoteAddress, FlexibleRollout
from UnleashClient.constants import METRIC_LAST_SENT_TIME, DISABLED_VARIATION
from .utils import LOGGER
from .deprecation_warnings import strategy_v2xx_deprecation_check


# pylint: disable=dangerous-default-value
class UnleashClient():
    """
    Client implementation.

    """
    def __init__(self,
                 url: str,
                 app_name: str,
                 environment: str = "default",
                 instance_id: str = "unleash-client-python",
                 refresh_interval: int = 15,
                 metrics_interval: int = 60,
                 disable_metrics: bool = False,
                 disable_registration: bool = False,
                 custom_headers: dict = {},
                 custom_options: dict = {},
                 custom_strategies: dict = {},
                 cache_directory: str = None) -> None:
        """
        A client for the Unleash feature toggle system.

        :param url: URL of the unleash server, required.
        :param app_name: Name of the application using the unleash client, required.
        :param environment: Name of the environment using the unleash client, optinal & defaults to "default".
        :param instance_id: Unique identifier for unleash client instance, optional & defaults to "unleash-client-python"
        :param refresh_interval: Provisioning refresh interval in ms, optional & defaults to 15 seconds
        :param metrics_interval: Metrics refresh interval in ms, optional & defaults to 60 seconds
        :param disable_metrics: Disables sending metrics to unleash server, optional & defaults to false.
        :param custom_headers: Default headers to send to unleash server, optional & defaults to empty.
        :param custom_options: Default requests parameters, optional & defaults to empty.
        :param custom_strategies: Dictionary of custom strategy names : custom strategy objects
        :param cache_directory: Location of the cache directory. When unset, FCache will determine the location
        """
        # Configuration
        self.unleash_url = url.rstrip('\\')
        self.unleash_app_name = app_name
        self.unleash_environment = environment
        self.unleash_instance_id = instance_id
        self.unleash_refresh_interval = refresh_interval
        self.unleash_metrics_interval = metrics_interval
        self.unleash_disable_metrics = disable_metrics
        self.unleash_disable_registration = disable_registration
        self.unleash_custom_headers = custom_headers
        self.unleash_custom_options = custom_options
        self.unleash_static_context = {
            "appName": self.unleash_app_name,
            "environment": self.unleash_environment
        }

        # Class objects
        self.cache = FileCache(self.unleash_instance_id, app_cache_dir=cache_directory)
        self.features = {}  # type: Dict
        self.scheduler = BackgroundScheduler()
        self.fl_job = None  # type: Job
        self.metric_job = None  # type: Job
        self.cache[METRIC_LAST_SENT_TIME] = datetime.now(timezone.utc)
        self.cache.sync()

        # Mappings
        default_strategy_mapping = {
            "applicationHostname": ApplicationHostname,
            "default": Default,
            "gradualRolloutRandom": GradualRolloutRandom,
            "gradualRolloutSessionId": GradualRolloutSessionId,
            "gradualRolloutUserId": GradualRolloutUserId,
            "remoteAddress": RemoteAddress,
            "userWithId": UserWithId,
            "flexibleRollout": FlexibleRollout
        }

        if custom_strategies:
            strategy_v2xx_deprecation_check([x for x in custom_strategies.values()])  # pylint: disable=R1721

        self.strategy_mapping = {**custom_strategies, **default_strategy_mapping}

        # Client status
        self.is_initialized = False

    def initialize_client(self) -> None:
        """
        Initializes client and starts communication with central unleash server(s).

        This kicks off:
        * Client registration
        * Provisioning poll
        * Stats poll

        :return:
        """
        # Setup
        fl_args = {
            "url": self.unleash_url,
            "app_name": self.unleash_app_name,
            "instance_id": self.unleash_instance_id,
            "custom_headers": self.unleash_custom_headers,
            "custom_options": self.unleash_custom_options,
            "cache": self.cache,
            "features": self.features,
            "strategy_mapping": self.strategy_mapping
        }

        metrics_args = {
            "url": self.unleash_url,
            "app_name": self.unleash_app_name,
            "instance_id": self.unleash_instance_id,
            "custom_headers": self.unleash_custom_headers,
            "custom_options": self.unleash_custom_options,
            "features": self.features,
            "ondisk_cache": self.cache
        }

        # Register app
        if not self.unleash_disable_registration:
            register_client(self.unleash_url, self.unleash_app_name, self.unleash_instance_id,
                            self.unleash_metrics_interval, self.unleash_custom_headers,
                            self.unleash_custom_options, self.strategy_mapping)

        fetch_and_load_features(**fl_args)

        # Start periodic jobs
        self.scheduler.start()
        self.fl_job = self.scheduler.add_job(fetch_and_load_features,
                                             trigger=IntervalTrigger(seconds=int(self.unleash_refresh_interval)),
                                             kwargs=fl_args)

        if not self.unleash_disable_metrics:
            self.metric_job = self.scheduler.add_job(aggregate_and_send_metrics,
                                                     trigger=IntervalTrigger(seconds=int(self.unleash_metrics_interval)),
                                                     kwargs=metrics_args)

        self.is_initialized = True

    def destroy(self):
        """
        Gracefully shuts down the Unleash client by stopping jobs, stopping the scheduler, and deleting the cache.

        You shouldn't need this too much!

        :return:
        """
        self.fl_job.remove()
        if self.metric_job:
            self.metric_job.remove()
        self.scheduler.shutdown()
        self.cache.delete()

    @staticmethod
    def _get_fallback_value(default_value: bool, fallback_function: Callable, feature_name: str, context: dict) -> bool:
        if fallback_function:
            fallback_value = default_value or fallback_function(feature_name, context)
        else:
            fallback_value = default_value

        return fallback_value

    # pylint: disable=broad-except
    def is_enabled(self,
                   feature_name: str,
                   context: dict = {},
                   default_value: bool = False,
                   fallback_function: Callable = None) -> bool:
        """
        Checks if a feature toggle is enabled.

        Notes:
        * If client hasn't been initialized yet or an error occurs, flat will default to false.

        :param feature_name: Name of the feature
        :param context: Dictionary with context (e.g. IPs, email) for feature toggle.
        :param default_value: Allows override of default value.
        :param fallback_function: Allows users to provide a custom function to set default value.
        :return: True/False
        """
        context.update(self.unleash_static_context)

        if self.is_initialized:
            try:
                return self.features[feature_name].is_enabled(context, default_value)
            except Exception as excep:
                LOGGER.warning("Returning default value for feature: %s", feature_name)
                LOGGER.warning("Error checking feature flag: %s", excep)
                return self._get_fallback_value(default_value, fallback_function, feature_name, context)
        else:
            LOGGER.warning("Returning default value for feature: %s", feature_name)
            LOGGER.warning("Attempted to get feature_flag %s, but client wasn't initialized!", feature_name)
            return self._get_fallback_value(default_value, fallback_function, feature_name, context)


    # pylint: disable=broad-except
    def get_variant(self,
                    feature_name: str,
                    context: dict = {}) -> dict:
        """
        Checks if a feature toggle is enabled.  If so, return variant.

        Notes:
        * If client hasn't been initialized yet or an error occurs, flat will default to false.

        :param feature_name: Name of the feature
        :param context: Dictionary with context (e.g. IPs, email) for feature toggle.
        :return: Dict with variant and feature flag status.
        """
        context.update(self.unleash_static_context)

        if self.is_initialized:
            try:
                return self.features[feature_name].get_variant(context)
            except Exception as excep:
                LOGGER.warning("Returning default flag/variation for feature: %s", feature_name)
                LOGGER.warning("Error checking feature flag variant: %s", excep)
                return DISABLED_VARIATION
        else:
            LOGGER.warning("Returning default flag/variation for feature: %s", feature_name)
            LOGGER.warning("Attempted to get feature flag/variation %s, but client wasn't initialized!", feature_name)
            return DISABLED_VARIATION



import pytest
import uuid
from fcache.cache import FileCache
from UnleashClient.constants import FEATURES_URL
from tests.utilities.mocks import MOCK_ALL_FEATURES, MOCK_CUSTOM_STRATEGY


@pytest.fixture()
def cache_empty():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    yield temporary_cache
    temporary_cache.delete()


@pytest.fixture()
def cache_full():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache[FEATURES_URL] = MOCK_ALL_FEATURES
    temporary_cache.sync()
    yield temporary_cache
    temporary_cache.delete()


@pytest.fixture()
def cache_custom():
    cache_name = 'pytest_%s' % uuid.uuid4()
    temporary_cache = FileCache(cache_name)
    temporary_cache[FEATURES_URL] = MOCK_CUSTOM_STRATEGY
    temporary_cache.sync()
    yield temporary_cache
    temporary_cache.delete()



import time
import json
import pytest
import responses
from UnleashClient import UnleashClient
from UnleashClient.strategies import Strategy
from tests.utilities.testing_constants import URL, ENVIRONMENT, APP_NAME, INSTANCE_ID, REFRESH_INTERVAL, \
    METRICS_INTERVAL, DISABLE_METRICS, DISABLE_REGISTRATION, CUSTOM_HEADERS, CUSTOM_OPTIONS
from tests.utilities.mocks.mock_features import MOCK_FEATURE_RESPONSE
from tests.utilities.mocks.mock_all_features import MOCK_ALL_FEATURES
from UnleashClient.constants import REGISTER_URL, FEATURES_URL, METRICS_URL


class EnvironmentStrategy(Strategy):
    def load_provisioning(self) -> list:
        return [x.strip() for x in self.parameters["environments"].split(',')]

    def apply(self, context: dict = None) -> bool:
        """
        Turn on if environemnt is a match.

        :return:
        """
        default_value = False

        if "environment" in context.keys():
            default_value = context["environment"] in self.parsed_provisioning

        return default_value


@pytest.fixture()
def unleash_client(tmpdir):
    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        refresh_interval=REFRESH_INTERVAL,
        metrics_interval=METRICS_INTERVAL,
        cache_directory=tmpdir.dirname
    )
    yield unleash_client
    unleash_client.destroy()


@pytest.fixture()
def unleash_client_nodestroy(tmpdir):
    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        refresh_interval=REFRESH_INTERVAL,
        metrics_interval=METRICS_INTERVAL,
        cache_directory=tmpdir.dirname
    )
    yield unleash_client


@pytest.fixture()
def unleash_client_toggle_only(tmpdir):
    unleash_client = UnleashClient(
        URL,
        APP_NAME,
        refresh_interval=REFRESH_INTERVAL,
        metrics_interval=METRICS_INTERVAL,
        disable_registration=True,
        disable_metrics=True,
        cache_directory=str(tmpdir)
    )
    yield unleash_client
    unleash_client.destroy()


def test_UC_initialize_default():
    client = UnleashClient(URL, APP_NAME)
    assert client.unleash_url == URL
    assert client.unleash_app_name == APP_NAME
    assert client.unleash_metrics_interval == 60


def test_UC_initialize_full():
    client = UnleashClient(URL,
                           APP_NAME,
                           ENVIRONMENT,
                           INSTANCE_ID,
                           REFRESH_INTERVAL,
                           METRICS_INTERVAL,
                           DISABLE_METRICS,
                           DISABLE_REGISTRATION,
                           CUSTOM_HEADERS,
                           CUSTOM_OPTIONS)
    assert client.unleash_instance_id == INSTANCE_ID
    assert client.unleash_refresh_interval == REFRESH_INTERVAL
    assert client.unleash_metrics_interval == METRICS_INTERVAL
    assert client.unleash_disable_metrics == DISABLE_METRICS
    assert client.unleash_disable_registration == DISABLE_REGISTRATION
    assert client.unleash_custom_headers == CUSTOM_HEADERS
    assert client.unleash_custom_options == CUSTOM_OPTIONS


def test_UC_type_violation():
    client = UnleashClient(URL, APP_NAME, refresh_interval="60")
    assert client.unleash_url == URL
    assert client.unleash_app_name == APP_NAME
    assert client.unleash_refresh_interval == "60"


@responses.activate
def test_uc_lifecycle(unleash_client):
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(1)
    assert unleash_client.is_initialized
    assert len(unleash_client.features) >= 4

    # Simulate server provisioning change
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_ALL_FEATURES, status=200)
    time.sleep(30)
    assert len(unleash_client.features) >= 9


@responses.activate
def test_uc_is_enabled(unleash_client):
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(1)
    assert unleash_client.is_enabled("testFlag")


@responses.activate
def test_uc_fallbackfunction(unleash_client, mocker):
    def good_fallback(feature_name: str, context: dict) -> bool:
        return True

    def bad_fallback(feature_name: str, context: dict) -> bool:
        return False

    def context_fallback(feature_name: str, context: dict) -> bool:
        return context['wat']

    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)
    fallback_spy = mocker.Mock(wraps=good_fallback)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(1)
    # Non-existent feature flag, fallback_function
    assert unleash_client.is_enabled("notFoundTestFlag", fallback_function=fallback_spy)
    assert fallback_spy.call_count == 1
    fallback_spy.reset_mock()

    # Non-existent feature flag, default value, fallback_function
    assert unleash_client.is_enabled("notFoundTestFlag", default_value=True, fallback_function=bad_fallback)
    assert fallback_spy.call_count == 0

    # Existent feature flag, fallback_function
    assert unleash_client.is_enabled("testFlag", fallback_function=good_fallback)
    assert fallback_spy.call_count == 0


@responses.activate
def test_uc_dirty_cache(unleash_client_nodestroy):
    unleash_client = unleash_client_nodestroy
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(5)
    assert unleash_client.is_enabled("testFlag")
    unleash_client.scheduler.shutdown()

    # Check that everything works if previous cache exists.
    unleash_client.initialize_client()
    time.sleep(5)
    assert unleash_client.is_enabled("testFlag")


@responses.activate
def test_uc_is_enabled_with_context():
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    custom_strategies_dict = {
        "custom-context": EnvironmentStrategy
    }

    unleash_client = UnleashClient(URL, APP_NAME, environment='prod', custom_strategies=custom_strategies_dict)
    # Create Unleash client and check initial load
    unleash_client.initialize_client()

    time.sleep(1)
    assert unleash_client.is_enabled("testContextFlag")
    unleash_client.destroy()


@responses.activate
def test_uc_is_enabled_error_states(unleash_client):
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(1)
    assert not unleash_client.is_enabled("ThisFlagDoesn'tExist")
    assert unleash_client.is_enabled("ThisFlagDoesn'tExist", default_value=True)


@responses.activate
def test_uc_not_initialized_isenabled():
    unleash_client = UnleashClient(URL, APP_NAME)
    assert not unleash_client.is_enabled("ThisFlagDoesn'tExist")
    assert unleash_client.is_enabled("ThisFlagDoesn'tExist", default_value=True)


@responses.activate
def test_uc_get_variant():
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    unleash_client = UnleashClient(URL, APP_NAME)
    # Create Unleash client and check initial load
    unleash_client.initialize_client()

    time.sleep(1)
    # If feature flag is on.
    variant = unleash_client.get_variant("testVariations", context={'userId': '2'})
    assert variant['name'] == 'VarA'
    assert variant['enabled']

    # If feature flag is not.
    variant = unleash_client.get_variant("testVariations", context={'userId': '3'})
    assert variant['name'] == 'disabled'
    assert not variant['enabled']

    unleash_client.destroy()


@responses.activate
def test_uc_not_initialized_getvariant():
    unleash_client = UnleashClient(URL, APP_NAME)
    variant = unleash_client.get_variant("ThisFlagDoesn'tExist")
    assert not variant['enabled']
    assert variant['name'] == 'disabled'


@responses.activate
def test_uc_metrics(unleash_client):
    # Set up API
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=202)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=202)

    # Create Unleash client and check initial load
    unleash_client.initialize_client()
    time.sleep(1)
    assert unleash_client.is_enabled("testFlag")

    time.sleep(12)
    request = json.loads(responses.calls[-1].request.body)
    assert request['bucket']["toggles"]["testFlag"]["yes"] == 1


@responses.activate
def test_uc_disabled_registration(unleash_client_toggle_only):
    unleash_client = unleash_client_toggle_only
    # Set up APIs
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=401)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=401)

    unleash_client.initialize_client()
    unleash_client.is_enabled("testFlag")
    time.sleep(20)
    assert unleash_client.is_enabled("testFlag")

    for api_call in responses.calls:
        assert '/api/client/features' in api_call.request.url


@responses.activate
def test_uc_server_error(unleash_client):
    # Verify that Unleash Client will still fall back gracefully if SERVER ANGRY RAWR, and then recover gracefully.

    unleash_client = unleash_client
    # Set up APIs
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=401)
    responses.add(responses.GET, URL + FEATURES_URL, status=500)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=401)

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("testFlag")

    responses.remove(responses.GET, URL + FEATURES_URL)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    time.sleep(20)
    assert unleash_client.is_enabled("testFlag")


@responses.activate
def test_uc_server_error_recovery(unleash_client):
    # Verify that Unleash Client will still fall back gracefully if SERVER ANGRY RAWR, and then recover gracefully.

    unleash_client = unleash_client
    # Set up APIs
    responses.add(responses.POST, URL + REGISTER_URL, json={}, status=401)
    responses.add(responses.GET, URL + FEATURES_URL, status=500)
    responses.add(responses.POST, URL + METRICS_URL, json={}, status=401)

    unleash_client.initialize_client()
    assert not unleash_client.is_enabled("testFlag")

    responses.remove(responses.GET, URL + FEATURES_URL)
    responses.add(responses.GET, URL + FEATURES_URL, json=MOCK_FEATURE_RESPONSE, status=200)
    time.sleep(20)
    assert unleash_client.is_enabled("testFlag")



from UnleashClient.variants import Variants
from UnleashClient.utils import LOGGER
from UnleashClient.constants import DISABLED_VARIATION


# pylint: disable=dangerous-default-value, broad-except
class Feature:
    def __init__(self,
                 name: str,
                 enabled: bool,
                 strategies: list,
                 variants: Variants = None) -> None:
        """
        An representation of a fewature object

        :param name: Name of the feature.
        :param enabled: Whether feature is enabled.
        :param strategies: List of sub-classed Strategy objects representing feature strategies.
        """
        # Experiment information
        self.name = name
        self.enabled = enabled
        self.strategies = strategies
        self.variations = variants

        # Stats tracking
        self.yes_count = 0
        self.no_count = 0

    def reset_stats(self) -> None:
        """
        Resets stats after metrics reporting

        :return:
        """
        self.yes_count = 0
        self.no_count = 0

    def increment_stats(self, result: bool) -> None:
        """
        Increments stats.

        :param result:
        :return:
        """
        if result:
            self.yes_count += 1
        else:
            self.no_count += 1

    def is_enabled(self,
                   context: dict = None,
                   default_value: bool = False) -> bool:
        """
        Checks if feature is enabled.

        :param context: Context information
        :param default_value: Optional, but allows for override.
        :return:
        """
        flag_value = default_value

        if self.enabled:
            try:
                if self.strategies:
                    strategy_result = any([x.execute(context) for x in self.strategies])
                else:
                    # If no strategies are present, should default to true.  This isn't possible via UI.
                    strategy_result = True

                flag_value = flag_value or strategy_result
            except Exception as strategy_except:
                LOGGER.warning("Error checking feature flag: %s", strategy_except)

        self.increment_stats(flag_value)

        LOGGER.info("Feature toggle status for feature %s: %s", self.name, flag_value)

        return flag_value

    def get_variant(self,
                    context: dict = None) -> dict:
        """
        Checks if feature is enabled and, if so, get the variant.

        :param context: Context information
        :return:
        """
        is_feature_enabled = self.is_enabled(context)

        if is_feature_enabled and self.variations is not None:
            try:
                variant = self.variations.get_variant(context)
                variant['enabled'] = is_feature_enabled
            except Exception as variant_exception:
                LOGGER.warning("Error selecting variant: %s", variant_exception)
        else:
            variant = DISABLED_VARIATION

        return variant
